#include "catch.hpp"
#include "expr.hpp"
#include "tests-small6-data.hpp"

#include <tuple>

static const double factor = 0.0001;

TEST_CASE("Edge cases") {
    std::string input;
    double expected;
    SECTION("Singular values") {
        input = "1.23";
        expected = 1.23;
        REQUIRE(create_expression_tree(input)->evaluate() == Approx(expected));
        input = "(1.23)";
        expected = 1.23;
        REQUIRE(create_expression_tree(input)->evaluate() == Approx(expected));
    }
    SECTION("Overused parens") {
        input = "((1.12)+(2.23))";
        expected = 3.35;
        REQUIRE(create_expression_tree(input)->evaluate() == Approx(expected));
        input = "(((1.12)+(2.23))-(2.3))";
        expected = 1.05;
        REQUIRE(create_expression_tree(input)->evaluate() == Approx(expected));
    }
    SECTION("Associativity") {
        input = "2^3 ^ 2";
        expected = 512;
        REQUIRE(create_expression_tree(input)->evaluate() == Approx(expected));
    }
}

TEST_CASE("Operator basics") {
    for (const auto& test : simple_tests) {
        std::string input;
        double expected;
        for (const auto& c : test) {
            std::tie(input, expected) = c;
            INFO("For input: " << input);
            auto e = create_expression_tree(input);
            INFO("Got expression: " << *e);
            REQUIRE(e->evaluate() == Approx(expected).epsilon(std::min(fabs(expected*factor), 1.)));
        }
    }
}

TEST_CASE("Complex Expressions") {
    std::string input;
    double expected;
    for (const auto& c : complex_tests) {
        std::tie(input, expected) = c;
        INFO("For input: " << input);
        auto e = create_expression_tree(input);
        INFO("Got expression: " << *e);
        REQUIRE(e->evaluate() == Approx(expected).epsilon(std::min(fabs(expected*factor), 1.)));
    }
}
